import requests
from requests.auth import HTTPBasicAuth
from .config import AppConfig
from .time_entry import TimeEntry
from .exceptions import TogglCommunicationException, TogglCredentialsException


class TogglIntegrator():
    """ Used to integrate with Toggl and get time entries """

    reports_url = 'https://toggl.com/reports/api/v2/details?'
    time_entries_url = 'https://www.toggl.com/api/v8/time_entries/'

    def __init__(self, config):
        self.config = config
        self.token = config.get('toggl_api_token')

    def fetch_time_entries(self):
        url = self.build_fetch_url()
        try:
            response = requests.get(
                url, auth=HTTPBasicAuth(self.token, 'api_token'))
        except ConnectionError:
            raise TogglCommunicationException()

        if not response.ok:
            raise TogglCredentialsException(response.json())

        time_entries = []
        for time_entry_data in response.json()['data']:
            time_entries.append(TimeEntry(time_entry_data))
        return time_entries
    
    def build_fetch_url(self):
        return TogglIntegrator.reports_url + self.get_time_entries_params()
    
    def update_saved_entries(self, entries):
        url = self.build_update_tags_url(entries)
        tag_info = {
            "time_entry": {
                "tags": ["SAVED"],
                "tag_action": "add"
            }
        }
        requests.put(
            url,
            data=tag_info,
            auth=HTTPBasicAuth(self.token, 'api_token')
        )
    
    def build_update_tags_url(self, entries):
        entries_ids = list(map(lambda e: e.id, entries))
        base_url = TogglIntegrator.time_entries_url
        ids = ','.join(entries_ids)
        return base_url + ids

    def get_time_entries_params(self):
        params = {
            "user_agent": self.config.get('toggl_user'),
            "workspace_id": self.config.get('toggl_workspace_id'),
            "since": self.config.get('toggl_start_date'),
            "until": self.config.get('toggl_end_date')
        }

        params_array = []
        for key in params:
            params_array.append(f'{key}={params[key]}')

        return '&'.join(params_array)



from .exceptions import ConfigException
from datetime import date


class AppConfig():

    toggl_config_keys = ['toggl_user', 'toggl_workspace_id',
                         'toggl_start_date', 'toggl_end_date', 'toggl_api_token']

    def __init__(self, config):
        self.config = config
        self.set_date_interval(config)
        AppConfig.__check_config_integrity(self.config)

    def set_date_interval(self, config):
        dates = ['toggl_start_date', 'toggl_end_date']
        for date_config in dates:
            if date_config in config:
                self.config[date_config] = config[date_config]
            else:
                self.config[date_config] = str(date.today())

    def get(self, config_name):
        return self.config.get(config_name)

    def set_start_date(self, date: str):
        self.config['toggl_start_date'] = date

    def toggl_config(self):
        return self.__get_config_from_keys(AppConfig.toggl_config_keys)

    def __get_config_from_keys(self, keys: list):
        config = {}
        for key in keys:
            config[key] = self.get(key)
        return config

    @staticmethod
    def __check_config_integrity(config):
        config_keys = ['toggl_user', 'toggl_workspace_id', 'toggl_start_date',
                       'toggl_end_date', 'toggl_api_token']
        for key in config_keys:
            if key not in config:
                raise ConfigException(
                    "There's a key missing in the configuration")



import logging
from .config import AppConfig
from .exceptions import ConfigException, EntryNotImportedException
from .time_entry import TimeEntry
from .toggl_integrator import TogglIntegrator

logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s', level=logging.INFO)


def toggl_import(config):
    """ A program to automaticaly import Toggl time entries """
    app_config = AppConfig(config)
    toggl = TogglIntegrator(app_config.toggl_config())

    time_entries = toggl.fetch_time_entries()

    entries_logged = []
    entries_not_logged = []

    import_handler = config['import_handler']

    for time_entry in time_entries:
        if time_entry.should_be_imported():
            try:
                import_handler(time_entry)
                entries_logged.append(time_entry)
            except EntryNotImportedException as e:
                msg = str(e)
                if msg == None or msg == '':
                    raise "There was no message explaining import failure"
                time_entry.message = msg
                entries_not_logged.append(time_entry)
        else:
            entries_not_logged.append(time_entry)
    
    toggl.update_saved_entries(entries_logged)

    return {
        'entries_saved': entries_logged,
        'entries_not_saved': entries_not_logged
    }



class ConfigException(Exception):
    """ Exception used when the stored config is corrupted """
    pass


class TimeEntryValidationError(Exception):
    """ Exception used to express a validation problem on Time Entry """
    pass


class TogglCommunicationException(Exception):
    """ Exception used when the communication with Toggl Failed """
    pass


class TogglCredentialsException(Exception):
    """ Exception used when Toggl rejected user credentials """
    pass

class EntryNotImportedException(Exception):
    """
    Use this exception when you want Toggl Import to know that the current 
    entry was not imported by your custom logic. Be sure to add a message
    when using this exception, as it should be reported to the user as the reason
    why the entry was not imported. 
    """
    pass


from datetime import datetime
from decimal import Decimal
from .exceptions import TimeEntryValidationError

activities = {'Design': 8,
              'Development': 9,
              'Analysis': 11,
              'Management': 15,
              'Documentation': 16,
              'Testing': 17,
              'Meeting': 18,
              'Training': 19,
              'Infrastructure': 20}


class TimeEntry():
    """ It represents a time entry object on Redmine """

    def __init__(self, time_entry_data):
        self.data = time_entry_data
        self.set_toggl_fields()

        if self.__should_be_ignored():
            self.ignore_on_import = True
            self.message = 'Entry ignored'
        else:
            self.set_import_fields()

    def set_toggl_fields(self):
        self.set_id()
        self.set_start_date()
        self.set_worked_hours()
        self.set_comment()
        self.set_timestamp()

    def __should_be_ignored(self):
        ignore = False
        flags = ['IGNORE', 'SAVED', 'LOGGED']
        for tag in self.data['tags']:
            tag = tag.strip().upper()
            if tag in flags:
                ignore = True

        return ignore

    def set_import_fields(self):
        try:
            self.validate_tags()
            self.set_issue_id()
            self.set_activity()
            self.ignore_on_import = False
        except TimeEntryValidationError as e:
            self.ignore_on_import = True
            self.message = str(e)

    def validate_tags(self):
        if 'tags' not in self.data or len(self.data['tags']) < 2:
            raise TimeEntryValidationError('Entry tags not set properly.')

    def set_id(self):
        self.id = self.data['id']

    def set_issue_id(self):
        self.issue_id = self.data['tags'][0].strip()
        try:
            int(self.issue_id)
        except ValueError:
            raise TimeEntryValidationError('Issue id is not an integer.')

    def set_activity(self):
        self.activity = self.data['tags'][1].strip()
        if self.activity not in activities:
            raise TimeEntryValidationError(
                'Activity is not in the possible values.')

    def set_start_date(self):
        start_date_time = datetime.fromisoformat(self.data['start'])
        self.start_date = start_date_time.date().isoformat()

    def set_worked_hours(self):
        hours_in_millis = Decimal(self.data['dur'])
        hours_in_seconds = hours_in_millis / 1000
        hours = hours_in_seconds / 3600
        self.worked_hours = float(round(hours, 2))

    def set_comment(self):
        self.comment = self.data['description']

    def set_timestamp(self):
        hours_in_millis = self.data['dur']
        hours_in_seconds = hours_in_millis / 1000
        hours = int(hours_in_seconds // 3600)
        remaining_seconds = hours_in_seconds - hours * 3600
        minutes = int(remaining_seconds // 60)
        seconds = int(remaining_seconds - minutes * 60)
        self.timestamp = f"{hours}:{minutes}:{seconds}"

    def should_be_imported(self):
        return not self.ignore_on_import

    def log(self):
        return f'Issue Id: {self.issue_id} | Data: {self.start_date} | Hora: {self.worked_hours} | Comment: {self.comment} | Activity: {self.activity}'
